#ifdef DEBUG
#include <cstdio>
#include <cstdlib>
#include <cstring>
#endif // DEBUG

#include <FL/Fl.H>
#include <FL/Fl_Double_Window.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_RGB_Image.H>
#include <FL/fl_draw.H>

#include <vector>
#include <algorithm>


////////////////////////////////////////////////////////////////////////////////
// Some conflic test and polygon algorithm refered to
// http://alienryderflex.com/polygon/
////////////////////////////////////////////////////////////////////////////////

#define NEARTEST_PX_SZ          10

#ifndef MIN
    #define MIN(a,b) (((a)<(b))?(a):(b))
#endif

#ifndef MAX
    #define MAX(a,b) (((a)>(b))?(a):(b))
#endif

////////////////////////////////////////////////////////////////////////////////

using namespace std;

////////////////////////////////////////////////////////////////////////////////

class TestDrawBox : public Fl_Box
{
    public:
        TestDrawBox(int x,int y,int w,int h);
        virtual ~TestDrawBox();

    protected:
        int handle( int event );
        void draw();

    public:
        typedef struct
        { int x; int y; } vecponint;

    protected:
        vector< vecponint > points;

    protected:
        void createme();
        void reorderpoints();
        void updatevectorregion();
        bool checkconflict( vecponint* refvp );
        bool checkduplicated( vecponint* refvp, int* idx = NULL );
        void drawxmark( int x, int y, int brdr = 1, int dst = 2 );
        void drawcursor( int x, int y );
        bool checknearpoints( int* x, int* y );

    protected:
        Fl_RGB_Image*   imgme;
        uchar*          imgmebuffer;
        int             last_mx;
        int             last_my;
        bool            inassemble;
        bool            oncompleted;
        bool            drawmouse;
        bool            holddraw;

};

////////////////////////////////////////////////////////////////////////////////

bool sortcondition (int i,int j)
{
    return ( i < j );
}

////////////////////////////////////////////////////////////////////////////////

TestDrawBox::TestDrawBox(int x,int y,int w,int h)
 : Fl_Box( x, y, w, h ),
   imgme( NULL ),
   imgmebuffer( NULL ),
   inassemble( false ),
   oncompleted( false ),
   last_mx( 0 ),
   last_my( 0 ),
   holddraw( false ),
   drawmouse( false )
{
    createme();
}

TestDrawBox::~TestDrawBox()
{
}

void TestDrawBox::createme()
{
    imgmebuffer = new uchar[ w() * h() * 3 ];
    if ( imgmebuffer != NULL )
    {
        // Let fill in white color !
        memset( imgmebuffer, 0xFF, w() * h() * 3 );

        // Create a FL_RGB_Image.
        imgme = new Fl_RGB_Image( imgmebuffer, w(), h(), 3 );
    }
}

void TestDrawBox::reorderpoints()
{
    #define TestDrawBoxSwapPt( _a_, _b_ ) \
            { vecponint swp = { _b_.x, _b_.y }; \
              _b_.x = _a_.x; _b_.y = _a_.y; \
              _a_.x = swp.x; _a_.y = swp.y }

    int ptsz = points.size();

    if ( ptsz > 2 )
    {
        int idxFirst = -1;
        int minX = imgme->w();
        int minY = imgme->h();

        // Find smallest coordination.
        // Y first.
        for( int cnt=0; cnt<(ptsz - 1); cnt++ )
        {
            if ( points[cnt].y < minY )
            {
                minX = points[cnt].x;
                minY = points[cnt].y;
                idxFirst = cnt;

                printf("[Y-Order] minX = %d, minY = %d, idx = %d         \r",
                       minX, minY, idxFirst );
            }
        }

        printf("\n");

        // X is next.
        for( int cnt=0; cnt<(ptsz - 1); cnt++ )
        {
            if ( ( points[cnt].x < minX ) && ( points[cnt].y < minY ) )
            {
                minX = points[cnt].x;
                minY = points[cnt].y;
                idxFirst = cnt;

                printf("[X-Order] minX = %d, minY = %d, idx = %d         \r",
                       minX, minY, idxFirst );
            }
        }

        printf("\n");

        if ( idxFirst > 0 )
        {
            vector< vecponint > copydummy;
            copydummy.resize( points.size() );
            for( int cpcnt=0; cpcnt<points.size()-1; cpcnt++ )
            {
                copydummy[cpcnt].x = points[cpcnt].x;
                copydummy[cpcnt].y = points[cpcnt].y;
            }

            int lastQ = 0;

            // copy all except last coordination
            for( int cnt=idxFirst; cnt<(ptsz - 1); cnt++ )
            {
                points[cnt-idxFirst].x = copydummy[cnt].x;
                points[cnt-idxFirst].y = copydummy[cnt].y;

                lastQ = cnt - idxFirst;
            }

            // Last end point will be re-assign for new vector's first.

            for( int cnt=0; cnt<idxFirst; cnt++ )
            {
                points[ cnt + lastQ + 1 ].x = copydummy[ cnt ].x;
                points[ cnt + lastQ + 1 ].y = copydummy[ cnt ].y;
            }

            // make perfect ring vector. start pos = end pos.
            points[ points.size() - 1 ].x = points[ 0 ].x;
            points[ points.size() - 1 ].y = points[ 0 ].y;

            copydummy.clear();
        }
    }
}

void TestDrawBox::updatevectorregion()
{
    if ( imgme == NULL )
        return;

    if ( imgmebuffer == NULL )
        return;

    if ( oncompleted == true )
    {
        holddraw = true;

        reorderpoints();

        if ( points.size() < 2 )
        {
            holddraw = false;
            return;
        }

        const int max_y = imgme->h();
        const int max_x = imgme->w();

        vector< double > node_x;

        unsigned fillcount = 0;

        for( int cur_y=0; cur_y<max_y; cur_y++ )
        {
            int         ptsz        = points.size();
            int         rcnt        = ptsz - 1;
            unsigned    node_count  = 0;

            for( int cnt=0; cnt<ptsz; cnt++ )
            {
                if ( ( (double)points[cnt].y   < (double)cur_y ) &&
                     ( (double)points[rcnt].y >= (double)cur_y ) ||
                     ( (double)points[rcnt].y  < (double)cur_y ) &&
                     ( (double)points[cnt].y  >= (double)cur_y ) )
                {
                    double newv = (double)points[cnt].x +
                                  ( (double)cur_y - (double)points[cnt].y ) /
                                  ( (double)points[rcnt].y - (double)points[cnt].y ) *
                                  ( (double)points[rcnt].x - (double)points[cnt].x );

                    node_x.push_back( newv );
                }

                rcnt = cnt;
            }

            int node_x_sz = node_x.size();

            // sort nodes ..
            if ( node_x_sz > 1 )
            {
                sort( node_x.begin(),
                      node_x.begin() + node_x_sz,
                      sortcondition );

                printf("nodes[%04d] : ", cur_y );
                for( int ncnt=0; ncnt<node_x_sz; ncnt++ )
                {
                    printf("%.2f, ", node_x[ncnt] );
                }
                printf("\n");

            }

            for( int dcnt=0; dcnt<node_x_sz; dcnt+=2 )
            {
                if ( node_x[dcnt] >= max_x )
                    break;

                if ( node_x[dcnt+1] > 0 )
                {
                    if ( node_x[dcnt] < 0 )
                    {
                        node_x[dcnt] = 0;
                    }

                    if ( node_x[dcnt+1] > max_x )
                    {
                        node_x[dcnt+1] = max_x;
                    }

                    for( int cur_x=node_x[dcnt]; cur_x<=node_x[dcnt+1]; cur_x++ )
                    {
                        int buffpos = ( max_x * cur_y + cur_x ) * 3;
                        imgmebuffer[ buffpos + 0 ] = 0x33;
                        imgmebuffer[ buffpos + 1 ] = 0x33;
                        imgmebuffer[ buffpos + 2 ] = 0xFF;
                        fillcount++;
                    }
                }
                else /// for test ...
                {
                    for( int cur_x=node_x[dcnt]; cur_x<max_x; cur_x++ )
                    {
                        int buffpos = ( max_x * cur_y + cur_x ) * 3;
                        imgmebuffer[ buffpos + 0 ] = 0x33;
                        imgmebuffer[ buffpos + 1 ] = 0xFF;
                        imgmebuffer[ buffpos + 2 ] = 0xFF;
                    }
                }
            }

            node_x.clear();

        } /// of for( y .. )

        printf(" fill = %d pixels.\n", fillcount );
    }
    else
    {
        memset( imgmebuffer, 0xFF, w() * h() * 3 );
    }

    imgme->uncache();

    holddraw = false;

    redraw();
}

bool TestDrawBox::checkconflict( vecponint* refvp )
{
    int vpsz = points.size() - 1;

    if ( vpsz < 2 )
        return false;

    vector< double > constants;
    vector< double > multiples;

    int rcnt = vpsz - 1;

    // Pre-scale first,
    for( int cnt=0; cnt<vpsz; cnt++ )
    {
        if ( (double)points[rcnt].y == (double)points[cnt].y )
        {
            constants.push_back( (double)points[cnt].x );
            multiples.push_back( 0.0 );
        }
        else
        {
            double newc = (double)points[cnt].x -
                          ( (double)points[cnt].y  * (double)points[rcnt].x ) /
                          ( (double)points[rcnt].y - (double)points[cnt].y  ) +
                          ( (double)points[cnt].y  * (double)points[cnt].x  ) /
                          ( (double)points[rcnt].y - (double)points[cnt].y  );

            double newm = ( (double)points[rcnt].x - (double)points[cnt].x ) /
                          ( (double)points[rcnt].y - (double)points[cnt].y );

            constants.push_back( newc );
            multiples.push_back( newm );
        }

        rcnt = cnt;
    }

    // Check new point in polygon region ...
    bool oddNodes = false;
    rcnt = vpsz - 1;

    for( int cnt=0; cnt<vpsz; cnt++ )
    {
        if ( ( points[cnt].y < refvp->y ) &&
             ( points[rcnt].y >= refvp->y ) &&
             ( points[rcnt].y < refvp->y ) &&
             ( points[cnt].y >= refvp->y ) )
        {
            oddNodes ^= ( (double)refvp->y * multiples[cnt] * constants[cnt] < (double)refvp->x );
        }

        rcnt = cnt;
    }

    return oddNodes;
}

bool TestDrawBox::checkduplicated( vecponint* refvp, int* idx )
{
    int vpsz = points.size();

    if ( vpsz == 0 )
        return false;

    if ( refvp != NULL )
    {
        for( int cnt=0; cnt< vpsz; cnt++ )
        {
            if ( ( points[cnt].x == refvp->x ) && ( points[cnt].y == refvp->y ) )
            {
                if ( idx != NULL )
                {
                    *idx = cnt;
                }

                return true;
            }
        }
    }

    return false;
}

void TestDrawBox::drawxmark( int x, int y, int brdr, int dst )
{
    fl_line_style( 0, brdr, 0 );
    fl_line( x - dst, y - dst, x + dst, y + dst );
    fl_line( x + dst, y - dst, x - dst, y + dst );
}

void TestDrawBox::drawcursor( int x, int y )
{
    for( int cnt=0; cnt<2; cnt++ )
    {
        fl_line_style( 0, cnt+1, 0 );

        if ( cnt == 0 )
            fl_color( FL_BLACK );
        else
            fl_color( FL_WHITE );

        int margin = 5 + ( 1 - cnt );
        int innerm = 2 - cnt;

        fl_line( x - margin, y - margin, x - innerm, y - innerm );
        fl_line( x + margin, y - margin, x + innerm, y - innerm );
        fl_line( x - margin, y + margin, x - innerm, y + innerm );
        fl_line( x + margin, y + margin, x + innerm, y + innerm );
    }

    fl_line_style( 0, 1, 0 );

    static char cursortext[64] = {0};

    sprintf(cursortext, "[ %d,%d ]", x, y );

    fl_color( FL_BLACK );
    fl_draw( cursortext, x + 7, y + 7 );
    fl_draw( cursortext, x + 9, y + 7 );
    fl_draw( cursortext, x + 9, y + 9 );
    fl_draw( cursortext, x + 7, y + 9 );

    fl_color( FL_WHITE );
    fl_draw( cursortext, x + 8, y + 8 );
}

bool TestDrawBox::checknearpoints( int* x, int* y )
{
    int vpsz = points.size();

    if ( vpsz < 3 )
        return false;

    // counting without last.
    for( int cnt=0; cnt<vpsz-1; cnt++ )
    {
        int minx = points[cnt].x - NEARTEST_PX_SZ;
        int miny = points[cnt].y - NEARTEST_PX_SZ;
        int maxx = points[cnt].x + NEARTEST_PX_SZ;
        int maxy = points[cnt].y + NEARTEST_PX_SZ;

        if ( ( minx < *x ) && ( maxx > *x ) && ( miny < *y ) && ( maxy > *y ) )
        {
            *x = points[cnt].x;
            *y = points[cnt].y;
            return true;
        }
    }

    return false;
}

int TestDrawBox::handle( int event )
{
    int mouse_btn = 0;
    int mouse_x = 0;
    int mouse_y = 0;

    switch( event )
    {
        case FL_FOCUS:
        case FL_UNFOCUS:
            return 1;

        case FL_ENTER:
            drawmouse = true;
            return 1;

        case FL_LEAVE:
            drawmouse = false;
            redraw();
            return 1;

        case FL_MOVE:
            last_mx = Fl::event_x();
            last_my = Fl::event_y();
            redraw();
            return 1;

        case FL_RELEASE:
            mouse_btn = Fl::event_button();
            if ( mouse_btn == FL_LEFT_MOUSE )
            {
                vecponint newvp {0};
                newvp.x = last_mx;
                newvp.y = last_my;

                int retidx = 0;

                if ( oncompleted == false )
                {
                    if ( checkduplicated( &newvp, &retidx ) == false )
                    {
                        points.push_back( newvp );

                        if ( inassemble == false )
                        {
                            inassemble = true;
                        }

                        int test_x = last_mx;
                        int test_y = last_my;

                        if ( checknearpoints( &test_x, &test_y ) == true )
                        {
                            points[ points.size() - 1 ].x = test_x;
                            points[ points.size() - 1 ].y = test_y;

                            if ( oncompleted == false )
                            {
                                oncompleted = true;
                                updatevectorregion();
                            }
                        }
                    }
                    else
                    {
                        // it's done to make region.
                        if ( ( retidx == 0 ) || ( retidx == points.size() - 1 ) )
                        {
                            newvp.x = points[0].x;
                            newvp.y = points[0].y;

                            points.push_back( newvp );

                            if ( oncompleted == false )
                            {
                                oncompleted = true;
                                updatevectorregion();
                            }
                        }
                    }
                }

                redraw();

                return 1;
            }
            else
            if ( mouse_btn == FL_RIGHT_MOUSE )
            {
                if ( oncompleted == true )
                {
                    // remove all vector points.
                    points.clear();
                    inassemble = false;
                    oncompleted = false;

                    updatevectorregion();
                    redraw();
                }
                else
                if ( inassemble == true )
                {
                    if ( points.size() > 0 )
                    {
                        // remove last one.
                        points.erase( points.begin() + points.size() - 1 );

                        if ( points.size() == 0 )
                        {
                            inassemble = false;
                        }

                        redraw();
                    }
                }
                else
                {
                    updatevectorregion();
                }
            }
    }
}

void TestDrawBox::draw()
{
    fl_push_clip( x(), y(), w(), h() );
    Fl_Color prevC = fl_color();

    fl_color( FL_RED );
    fl_font( FL_COURIER, 12 );

    if ( ( imgme != NULL ) && ( holddraw == false ) )
    {
        imgme->draw( 0, 0 );
    }

    if ( inassemble == true )
    {
        int vsz = points.size();

        char outnumbstr[32] = {0};

        unsigned coordtxtcolor = 0xCC99CC00;

        if ( vsz > 0 )
        {
            drawxmark( points[0].x, points[0].y, 2 );

            if ( oncompleted == false )
            {
                sprintf( outnumbstr, "[%d]", 0 );
            }
            else
            {
                sprintf( outnumbstr, "[%d] + [%d]", 0, vsz - 1 );
            }

            fl_color( coordtxtcolor );
            fl_draw( outnumbstr, points[0].x + 5, points[0].y + 5 );
            fl_color( FL_BLACK );
            fl_draw( outnumbstr, points[0].x + 6, points[0].y + 6 );

        }

        fl_color( FL_RED );
        fl_line_style( 0, 1, 0 );

        if ( vsz > 1 )
        {
            for( int cnt=1; cnt<vsz; cnt++ )
            {
                fl_line( points[cnt-1].x, points[cnt-1].y,
                         points[cnt  ].x, points[cnt  ].y );

                drawxmark( points[cnt].x, points[cnt].y, 2 );

                if ( ( cnt < vsz - 1 ) || ( oncompleted == false ) )
                {
                    sprintf( outnumbstr, "[%d]", cnt );

                    fl_color( coordtxtcolor );
                    fl_draw( outnumbstr, points[cnt].x + 25, points[cnt].y + 5 );
                    fl_color( FL_BLACK );
                    fl_draw( outnumbstr, points[cnt].x + 26, points[cnt].y + 6 );
                }

                fl_color( FL_RED );
                fl_line_style( 0, 1, 0 );
            }
        }
    }

    if ( drawmouse == true )
    {
        drawcursor( last_mx, last_my );
    }

    fl_color( prevC );
    fl_pop_clip();
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

Fl_Double_Window*   window = NULL;
TestDrawBox*        testbox = NULL;

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

int main (int argc, char ** argv)
{
    int reti = 0 ;

    window = new Fl_Double_Window ( 600, 400, "Testing polyfill, day #3 - sort bug fixed" );
    if ( window != NULL )
    {
        window->begin();

        testbox = new TestDrawBox( 0, 0, window->w(), window->h() );

        window->end();
        window->show();

        fl_cursor( FL_CURSOR_NONE, 0, 0 );

        reti = Fl::run();
    }

    return reti;
}
